/*
 * cp40tcl.h --
 *
 *	Definitions for CP40 interface with Tcl.
 *
 * Copyright (c) 1999, Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#ifndef CP40TCL_H
#define CP40TCL_H 1
#include <tcl.h>
#include <sys/time.h>
#include "cp40.h"

/*---------------------------------------------------------------------------*/
/* TYPE DEFINITIONS */

/* The following structure is used as client data for CP40 instance. */
typedef struct cp40tcl_instance_data_s cp40tcl_instance_data_t;
struct cp40tcl_instance_data_s {
  Tcl_Interp *interp;     /* Tcl's interpretor in which the instance
			     was created. */
  Tcl_Command command;    /* Tcl's token for the instance command. */
  int connection;         /* Real time socket file identifier (if >=0,
			     Tcl_DeleteFileHandler must be called before
			     closing connection). */
  char *script;           /* Tcl script to evaluate (at global level) after
			     all (at least one) pending photon frames have
			     been read and processed by scheduled real-time
			     tasks. */
  struct timeval start;   /* Date at which socket client was started. */
  int nframes;            /* Total number of frames read since socket
			     client was started. */
  char *name;             /* Malloc'ed instance name. */
  cp40_instance_t *owner; /* for backward reference */
  gpl_hash_entry_t *entry;      /* Hash entry in instances hash table. */
};

/*---------------------------------------------------------------------------*/
/* PUBLIC ROUTINES */

extern int Cp40tcl_Init(Tcl_Interp *interp);
/*	Initialise CP40 + Tcl shell.  Return standard Tcl result.  In case
	of error, a message is left in INTERP->result. */

extern cp40tcl_instance_data_t *cp40tcl_new_instance_data(Tcl_Interp *interp,
							  const char *name);
/*	Allocate a new CP40-TCL instance by name and create the related
	command in INTERP.  Return NULL in case of error and left a message in
	INTERP->result.  It is an error to re-create an existing command or
	instance.  NAME can be "#auto" to create a unique name. */

extern void cp40tcl_delete_instance_data(cp40tcl_instance_data_t *data,
					 int update);
/*	Delete DATA, free all related resources (DATA can be NULL, in which
	case nothing is done). */

/*---------------------------------------------------------------------------*/
/* GETTING INSTANCE (MEMBER) BY NAME */

extern cp40tcl_instance_data_t *cp40tcl_get_instance_data(const char *name);
/*	Get CP40-TCL instance data by name.  Return NULL in case of error. */

extern cp40_instance_t *cp40tcl_get_instance_owner(const char *name);
/*	Same as cp40tcl_get_instance_data(NAME)->owner. Return NULL in
	case of error. */

extern cp40_task_t *cp40tcl_get_task(const char *name);
/*	Get CP40-TCL task by name, NAME should be in the form "DATA->TASK".
	Return NULL in case of error. */


#endif /* CP40TCL_H ---------------------------------------------------------*/
