/*---------------------------------------------------------------------------*/
/* INTERACTIVE GRAPHICS */
struct pgcursor { float x, y; char c; };

func pgcurs(x, y)
{
  x= is_void(x) ? 0.0F : float(x);
  y= is_void(y) ? 0.0F : float(y);
  ch= "1";
  if (! __pgcurs(x, y, ch)) error, "PGCURS failed";
  return pgcursor(x=x, y=y, c=*pointer(ch)(1));
}

func pgband(xref, yref, x, y, mode=)
/* DOCUMENT cursor= pgband(xref, yref);
            cursor= pgband(xref, yref, x, y);
   KEYWORDS
     MODE (default 0)
*/
{
  xref= is_void(xref) ? 0.0F : float(xref);
  yref= is_void(yref) ? 0.0F : float(yref);
  if (is_void(mode)) mode= 0;
  if ((posn= ! is_void(x))) {
    x= float(x); y= float(y);
  } else {
    x= xref; y= yref;
  }
  ch= "1";
  if (! __pgband(mode, posn, xref, yref, x, y, ch)) error, "PGBAND failed";
  return pgcursor(x=x, y=y, c=*pointer(ch)(1));
}

local pgbbuf, pgebuf; pgbbuf= __pgbbuf; pgebuf= __pebbuf;
/* DOCUMENT pgbbuf;
            pgebuf;
     Begin/end batch of output (buffer). */

/*---------------------------------------------------------------------------*/
func pgenv(xmin, xmax, ymin, ymax, just=, axis=)
/* DOCUMENT pgenv, xmin, xmax, ymin, ymax;
   KEYWORDS: JUST, AXIS. */
{
  if (is_void(just)) just= 0;
  if (is_void(axis)) axis= 0;
  __pgenv, xmin, xmax, ymin, ymax, just, axis;
}

local pgask; pgask= __pgask;
/* DOCUMENT pgask, flag;
     Control new page prompting. */

local pgsvp; pgsvp= __pgvsp;
/* DOCUMENT pgsvp, left, right, bottom, top;
     Set viewport (normalized device coordinates). */

local pgvstd; pgvstd= __pgvstd;
/* DOCUMENT pgvstd;
     Set standard (default) viewport. */

local pgswin; pgswin= __pgswin;
/* DOCUMENT pgswin, left, right, bottom, top;
     Set viewport (normalized device coordinates). */

local pgwnad; pgwnad= __pgwnad;
/* DOCUMENT pgwnad, left, right, bottom, top;
     Set window and adjust viewport to same aspect ratio. */

func pg_viewport(x1, x2, y1, y2, units=)
/* DOCUMENT
   UNITS= 0 : normalized device coordinates (default)
   UNITS= 1 : inches
   UNITS= 2 : millimeters
   UNITS= 3 : pixels
   UNITS= 4 : data
 */
{
  if (is_void(units)) units= 0;
  if (is_void(x1)) {
    /* Query view size. */
    x1= x2= y1= y2= 0.0F;
    if (units==4) {
      __pgqwin, x1, x2, y1, y2;
    } else {
      __pgqvp, units, x1, x2, y1, y2;
    }
    return [x1,x2,y1,y2];
  }

  /* Set view size. */
  if (is_void(x2)) {
    if (numberof(x1)!=4) error, "BOX must be a four element vector";
    x2= x1(1); y1= x1(3); y2= x1(4); x1= x1(1);
  }
  if (units==4) {
    __pgswin, x1, x2, y1, y2;
  } else {
    __pgqvp, units, x1, x2, y1, y2;
  }
}

/*---------------------------------------------------------------------------*/
/* ATTRIBUTES */

local pgsave, pgunsa; pgsave= __pgsave; pgunsa= __pgunsa;
/* DOCUMENT pgsave;
            pgunsa;
     PGSAVE saves the current PGPLOT attributes in a private storage area.
     They can be restored by calling PGUNSA (unsave).  Attributes saved are:
     character font, character height, color index, fill-area style, line
     style, line width, pen position, arrow-head style, hatching style, and
     clipping state. Color representation is not saved.

     Calls to PGSAVE and PGUNSA should always be paired. Up to 20 copies of
     the attributes may be saved. PGUNSA always retrieves the last-saved
     values (last-in first-out stack).

     Note that when multiple devices are in use, PGUNSA retrieves the values
     saved by the last PGSAVE call, even if they were for a different device.*/

/*---------------------------------------------------------------------------*/
func pg_default(&symbol, value)
/* DOCUMENT pg_default, symbol, value;
       -or- value= pg_default(symbol, default_value);
     If called as a subroutine, set SYMBOL to VALUE if SYMBOL is void.
     Otherwise, returns SYMBOL if non-nil and DEFAULT_VALUE otherwise. */
{
  if (! am_subroutine()) return (is_void(symbol) ? value : symbol);
  if (is_void(symbol)) symbol= value;
}

func pg_reshape2(&x, &y)
/* DOCUMENT number= pg_reshape2(&x, &y);
     Force arrays X and Y having same dimensions and float type and return
     their number of elements. */
{
  if (is_void((dims= dimsof(x,y)))) error, "non-conformable arrays";
  number= numberof((zero= array(0.0f, dims)));
  if (numberof(x)!=number) x= float(x)+zero;
  if (numberof(y)!=number) y= float(y)+zero;
  return number;
}

/*---------------------------------------------------------------------------*/
/* PAGE SIZE */

func pg_pagesize(width, height, units=)
{
  pg_default, units, 1;  /* default units=inches */
  if (is_void(width)) {
    /* Query view size. */
    x1= x2= y1= y2= 0.0F;
    __pgqvz, units, x1, x2, y1, y2;
    return [(x2-x1),(y2-y1)];
  }

  /* Set view size. */
  if (is_void(height)) {
    n= numberof(width);
    if (n==1) {
      height= width;
    } else if (n==2) {
      height= width(2); width= height(1);
    } else {
      error, "bad WIDTH";
    }
  }
  if (units!=1) {
    /* Guess the UNITS->INCH scale factor. */
    x1= x2= y1= y2= 0.0F;
    __pgqvsz, 1, x1, x2, y1, y2; /* current page size in inches */
    wi= x2-x1; hi= y2-y1;
    __pgqvsz, units, x1, x2, y1, y2; /* current page size in units */
    wu= x2-x1; hu= y2-y1;
    width*= wi/wu;
    height*= hi/hu;
  }
  __pgpap, width, height/width;
}

/*---------------------------------------------------------------------------*/
/* OPENING, CLOSING, AND SELECTING DEVICES */

local pgopen; pgopen= _pgopen;
/* DOCUMENT id= pgopen(device);
     Open a graphics device for PGPLOT output. If the device is
     opened successfully, it becomes the selected device to which
     graphics output is directed until another device is selected
     with PGSLCT or the device is closed with PGCLOS.
   
     The value returned by PGOPEN should be tested to ensure that
     the device was opened successfully, e.g.,
   
     ISTAT = PGOPEN('plot.ps/PS')
     IF (ISTAT .LE. 0 ) STOP */

local pgclos; pgclos= __pgclos;
/* DOCUMENT pgclos;
     Close current PGPLOT device. */
local pgslct; pgslct= __pgslct;
/* DOCUMENT pgslct, id;
     Select an open graphics device. */

func pgqid(dummy)
/* DOCUMENT id= pgqid();
     Inquire current device identifier. */
{ id= 0n; __pgqid, id; return id; }

local pgpage; pgpage= __pgpage;
/* DOCUMENT pgpage;
     Advance to new page. */

local pgbox; pgbox= __pgbox;
/* PROTOTYPE void cpgbox(string xopt, float xtick, int nxsub, string yopt,
                         float ytick, int nysub); */

/*---------------------------------------------------------------------------*/
/* COLORS */

func pgqcol(which)
/* DOCUMENT [ci_min, ci_max]= pgqcol();
             ci_min= pgqcol(1);
             ci_max= pgqcol(2);
     Inquire color capability: CI_MIN is the minimum available color index.
     This will be either 0 if the device can write in the background color,
     or 1 if not.  CI_MAX is the maximum available color index.  This will
     be 1 if the device has no color capability, or a larger number
     (e.g., 3, 7, 15, 255). */
{
  ci1= ci2= 0n;
  __pgqcol, ci1, ci2;
  return [ci1,ci2](which);
}

func pgqcir(which)
/* DOCUMENT [ci_min, ci_max]= pgqcir();
             ci_min= pgqcir(1);
             ci_max= pgqcir(2);
     Inquire color capability: CI_MIN (CI_MAX) is the lowest (highest) color
     index to use for images.

   SEE ALSO: pgscir. */
{
  ci1= ci2= 0n;
  __pgqcol, ci1, ci2;
  return [ci1,ci2](which);
}

local pgscir; pgscir= __pgscir;
/* DOCUMENT pgscir, ci_min, ci_max;
     Set color index range for images.
   SEE ALSO: pgqcir. */

local pgsci; pgsci= __pgsci;
/* DOCUMENT pgsci, ci;
     Set color index for subsequent line/text draw. */

func pgqci(void)
/* DOCUMENT ci= pgqci();
     Query current color index. */
{ ci= 0n; __pgqci, ci; return ci; }

/*---------------------------------------------------------------------------*/
/* POLYGONS AND FILLED AREAS */

func pgpoly(x, y)
{
  n= pg_reshape2(x, y);
  __pgpoly, n, x, y;
}

local pgrect; pgrect= __pgrect;
/* DOCUMENT pgrect, x1, x2, y1, y2;
     Draw a rectangle, using fill-area attributes. */

local pgcirc; pgcirc= __pgcirc;
/* DOCUMENT pgcirc, xcent, ycent, radius;
     Draw a circle, using fill-area attributes. */

/*---------------------------------------------------------------------------*/
/* LINE PLOT */

func pgline(x,y)
{
  n= pg_reshape2(x, y);
  __pgline, n, x, y;
}

/*---------------------------------------------------------------------------*/
/* GRAPH MARKERS */

func pgpnt(x, y, s)
/* DOCUMENT pgpnt, x, y, s;
     Plot Y versus X with marker S. */
{
  nxy= pg_reshape2(x, y);
  ns= numberof(s);
  if (ns==1) __pgpt, nxy, x, y, s;
  else __pgpnts, nxy, x, y, s, ns;
}
/*---------------------------------------------------------------------------*/
/* DRAWING IMAGES */

func pgpixl(ia, x1, x2, y1, y2)
{
  if ((dims= dimsof(ia))(1)!=2) error, "expecting 2D array";
  idim= dims(2); jdim= dims(3);
  if (is_void(x1)) {
    x1= y1= 0.5f; x2= float(idim)+0.5f; y2= float(jdim)+0.5f;
  }
  __pgpixl, ia, idim, jdim, 1, idim, 1, jdim, x1, x2, y1, y2;
}

func pg_image(a)
{
  if (! pgqid()) pgopen, "/xwin";
  if ((dims= dimsof(a))(1)!=2) error, "expecting 2D array";
  idim= dims(2); jdim= dims(3);
  ci= pgqcir();
  cmin= ci(1);
  cmax= ci(2);
  amin= min(a);
  amax= max(a);
  x1= 0.5f;
  x2= idim+x1;
  y1= 0.5f;
  y2= jdim+y1;
  if (amax<=amin) return;
  if (amin) a-= amin;
  scale= ((cmax-cmin+1.0f)/(amax-amin));
  bias= cmin + 0.5f - scale*amin; 
  __pgenv, x1, x2, y1, y2, 0, 0;
  __pgpixl, int(scale*a + bias), idim, jdim, 1, idim, 1, jdim, x1, x2, y1, y2;
}

/*---------------------------------------------------------------------------*/
/* TEXT */

func pgqtxt(x, y, text, angle=, just=)
/* DOCUMENT bbox= pgqtxt(x, y, text);
     Find the bounding box for a text string as drawn by PGPTXT
     (arguments nad keywords are identical to PGPTXT).  The result is
     an array with the coordinates of the corners of a rectangle
     parallel to the string baseline that just encloses the string:
        BBOX= [[XLL, XUL, XUR, XUL], [YLL, YUL, YUR, YUL]]
     where ("left" and "right" refer to the first and last characters
     in the string):
        (XLL,YLL)= lower left corner
        (XUL,YUL)= upper left corner
        (XLR,YLR)= lower right corner
        (XUR,YUR)= upper right corner

     SEE ALSO: pgptxt. */
{
  xbox= ybox= array(float, 4);
  __pgqtxt, x, y, pg_default(angle, 0.0F), pg_default(just, 0.0F), text,
    xbox, ybox;
  return [xbox, ybox];
}

func pgptxt(x, y, text, angle=, just=)
/* DOCUMENT pgptxt, x, y, text;
     Write text at arbitrary position and orientation.  String TEXT is
     drawn at world coordinates (X,Y) using the current values of
     attributes color-index, line-width, character-height, and
     character-font.  Drawn text is NOT subject to clipping at the
     edge of the window.  Keyword ANGLE set the angle, in degrees,
     that the baseline is to make with the horizontal, increasing
     counter-clockwise (0.0 is horizontal and is the default angle).
     Keyword JUST controls horizontal justification of the string:
     with JUST=0.0 (the default) the string will be left-justified at
     the point (X,Y); centered with JUST=0.5, and right justified if
     JUST=1.0 (other values of JUST give other justifications).

   SEE ALSO: pgqtxt, pglen, pgtext. */
{
  __pgptxt, x, y, pg_default(angle, 0.0F), pg_default(just, 0.0F), text;
}

/* PROTOTYPE void cpgtext(float x, float y, const char *text); */

/* PROTOTYPE void cpgmtxt(const char *side, float disp, float coord, float fjust, const char *text); */

func pglen(units, text)
/* DOCUMENT [width, height]= pglen(units, text);
     Get dimensions of string TEXT in a variety of units:
     UNITS=  0 => answer in normalized device coordinates
             1 => answer in inches
             2 => answer in mm
             3 => answer in absolute device coordinates (dots)
             4 => answer in world coordinates
             5 => answer as a fraction of the current viewport size

   SEE ALSO: pgptxt, pgtext, pgqtxt. */
{ x= y= 0.0F;  _pglen, units, text, x, y; return [x,y]; }

local pgtext; pgtext= __pgtext;
/* DOCUMENT pgtext, x, y, text;
     Write TEXT (horizontal, left-justified) at world position (X,Y).
     The bottom left corner of the first character is placed
     at the specified position, and the text is written horizontally.
     This is a simplified interface to the primitive routine PGPTXT.
     For non-horizontal text, use PGPTXT.

   SEE ALSO: pgptxt. */

/*---------------------------------------------------------------------------*/

