/*
 * gpl_macros.h --
 *
 *	Macro definitions for manipulation of identifiers in GPLib library.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 */
#ifndef _GPL_MACROS_H
#define _GPL_MACROS_H 1
#include "gpl_base.h"

/* In the following macros, the ID argument is one of:
     BYTE,
     SHORT,
     INT,
     LONG,
     FLOAT,
     DOUBLE,
     FCOMPLEX
     DCOMPLEX.
   The defined macros are:
     GPL_IS_UNSIGNED(id)   is true if and only if ID is an unsigned integer;
     GPL_IS_INTEGER(id)    is true if and only if ID is an integer;
     GPL_IS_COMPLEX(id)    is true if and only if ID is a complex;
     GPL_IS_DOUBLE(id)     is true if and only if real part of ID is double
                         precision;
     TIM_ID_OF(id)         GPLib identifier for ID;
     GPL_CTYPE_OF(id)      C type for ID;
     GPL_REAL_PART_OF(id)  identifier for real part of ID (e.g. just ID if not
                         complex), for instance;
			   GPL_CTYPE_OF(GPL_REAL_PART_OF(DCOMPLEX))
                         gives `double'.
     GPL_ABBREV(id)        abbreviation letter for ID in GPLib;
     GPL_LOWER(id)         lower case name of ID in GPLib;
     GPL_CAPITALIZE(id)    capitalized name of ID in GPLib;
     GPL_PROMOTE(x,y)      identifier for promoted operand in a binary operation
                         between operands of identifier X and Y (should follow
			 C rules: GPL_PROMOTE(INT, DOUBLE) gives `DOUBLE'). */

#define GPL_IS_INTEGER(id)             GPL_JOIN(GPL_IS_INTEGER_, id)
#define GPL_IS_INTEGER_BYTE            1
#define GPL_IS_INTEGER_SHORT           1
#define GPL_IS_INTEGER_INT             1
#define GPL_IS_INTEGER_LONG            1
#define GPL_IS_INTEGER_FLOAT           0
#define GPL_IS_INTEGER_DOUBLE          0
#define GPL_IS_INTEGER_FCOMPLEX   0
#define GPL_IS_INTEGER_DCOMPLEX  0

#define GPL_IS_UNSIGNED(id)            GPL_JOIN(GPL_IS_UNSIGNED_, id)
#define GPL_IS_UNSIGNED_BYTE           1
#define GPL_IS_UNSIGNED_SHORT          0
#define GPL_IS_UNSIGNED_INT            0
#define GPL_IS_UNSIGNED_LONG           0
#define GPL_IS_UNSIGNED_FLOAT          0
#define GPL_IS_UNSIGNED_DOUBLE         0
#define GPL_IS_UNSIGNED_FCOMPLEX  0
#define GPL_IS_UNSIGNED_DCOMPLEX 0

#define GPL_IS_COMPLEX(id)             GPL_JOIN(GPL_IS_COMPLEX_, id)
#define GPL_IS_COMPLEX_BYTE            0
#define GPL_IS_COMPLEX_SHORT           0
#define GPL_IS_COMPLEX_INT             0
#define GPL_IS_COMPLEX_LONG            0
#define GPL_IS_FCOMPLEX           0
#define GPL_IS_DCOMPLEX          0
#define GPL_IS_COMPLEX_FCOMPLEX   1
#define GPL_IS_COMPLEX_DCOMPLEX  1

#define GPL_IS_DOUBLE(id)             GPL_JOIN(GPL_IS_DOUBLE_, id)
#define GPL_IS_DOUBLE_BYTE            0
#define GPL_IS_DOUBLE_SHORT           0
#define GPL_IS_DOUBLE_INT             0
#define GPL_IS_DOUBLE_LONG            0
#define GPL_IS_DOUBLE_FLOAT           0
#define GPL_IS_DOUBLE_DOUBLE          1
#define GPL_IS_DOUBLE_FCOMPLEX   0
#define GPL_IS_DOUBLE_DCOMPLEX  1

#define TIM_ID_OF(id)  GPL_JOIN(TIM_ID_, id)

#define GPL_CTYPE_OF(id)         GPL_JOIN(GPL_CTYPE_, id)
#define GPL_CTYPE_BYTE           gpl_byte_t
#define GPL_CTYPE_SHORT          short
#define GPL_CTYPE_INT            int
#define GPL_CTYPE_LONG           long
#define GPL_CTYPE_FLOAT          float
#define GPL_CTYPE_DOUBLE         double
#define GPL_CTYPE_FCOMPLEX  gpl_fcomplex_t
#define GPL_CTYPE_DCOMPLEX gpl_dcomplex_t

#define GPL_REAL_PART_OF(id)            GPL_JOIN(GPL_REAL_PART_OF_, id)
#define GPL_REAL_PART_OF_BYTE           BYTE
#define GPL_REAL_PART_OF_SHORT          SHORT
#define GPL_REAL_PART_OF_INT            INT
#define GPL_REAL_PART_OF_LONG           LONG
#define GPL_REAL_PART_OF_FLOAT          FLOAT
#define GPL_REAL_PART_OF_DOUBLE         DOUBLE
#define GPL_REAL_PART_OF_FCOMPLEX  FLOAT
#define GPL_REAL_PART_OF_DCOMPLEX DOUBLE

#define GPL_ABBREV(id)            GPL_JOIN(GPL_ABBREV_, id)
#define GPL_ABBREV_BYTE           b
#define GPL_ABBREV_SHORT          s
#define GPL_ABBREV_INT            i
#define GPL_ABBREV_LONG           l
#define GPL_ABBREV_FLOAT          f
#define GPL_ABBREV_DOUBLE         d
#define GPL_ABBREV_FCOMPLEX  c
#define GPL_ABBREV_DCOMPLEX z
#define GPL_ABBREV_STRING         q
#define GPL_ABBREV_POINTER        p
#define GPL_ABBREV_TYPE           t
#define GPL_ABBREV_VOID           v
#define GPL_ABBREV_ARRAY          a
#define GPL_ABBREV_OBJECT         o
#define GPL_ABBREV_FORBIDDEN      x

#define GPL_LOWER(id)            GPL_JOIN(GPL_LOWER_, id)
#define GPL_LOWER_BYTE           byte
#define GPL_LOWER_SHORT          short
#define GPL_LOWER_INT            int
#define GPL_LOWER_LONG           long
#define GPL_LOWER_FLOAT          float
#define GPL_LOWER_DOUBLE         double
#define GPL_LOWER_FCOMPLEX  fcomplex
#define GPL_LOWER_DCOMPLEX dcomplex

#define GPL_CAPITALIZE(id)            GPL_JOIN(GPL_CAPITALIZE_, id)
#define GPL_CAPITALIZE_BYTE           Byte
#define GPL_CAPITALIZE_SHORT          Short
#define GPL_CAPITALIZE_INT            Int
#define GPL_CAPITALIZE_LONG           Long
#define GPL_CAPITALIZE_FLOAT          Float
#define GPL_CAPITALIZE_DOUBLE         Double
#define GPL_CAPITALIZE_FCOMPLEX  ComplexFloat
#define GPL_CAPITALIZE_DCOMPLEX ComplexDouble


#define GPL_OPS_OF(id)   GPL_JOIN(timOps, GPL_CAPITALIZE(id))

#define GPL_PROMOTE(x,y) GPL_JOIN3(GPL_PROMOTE_, GPL_ABBREV(x), GPL_ABBREV(y)))

#define GPL_PROMOTE_BB   BYTE
#define GPL_PROMOTE_BS   SHORT
#define GPL_PROMOTE_BI   INT
#define GPL_PROMOTE_BL   LONG
#define GPL_PROMOTE_BF   FLOAT
#define GPL_PROMOTE_BD   DOUBLE
#define GPL_PROMOTE_BC   FCOMPLEX
#define GPL_PROMOTE_BZ   DCOMPLEX

#define GPL_PROMOTE_SB   SHORT
#define GPL_PROMOTE_SS   SHORT
#define GPL_PROMOTE_SI   INT
#define GPL_PROMOTE_SL   LONG
#define GPL_PROMOTE_SF   FLOAT
#define GPL_PROMOTE_SD   DOUBLE
#define GPL_PROMOTE_SC   FCOMPLEX
#define GPL_PROMOTE_SZ   DCOMPLEX

#define GPL_PROMOTE_IB   INT
#define GPL_PROMOTE_IS   INT
#define GPL_PROMOTE_II   INT
#define GPL_PROMOTE_IL   LONG
#define GPL_PROMOTE_IF   FLOAT
#define GPL_PROMOTE_ID   DOUBLE
#define GPL_PROMOTE_IC   FCOMPLEX
#define GPL_PROMOTE_IZ   DCOMPLEX

#define GPL_PROMOTE_LB   LONG
#define GPL_PROMOTE_LS   LONG
#define GPL_PROMOTE_LI   LONG
#define GPL_PROMOTE_LL   LONG
#define GPL_PROMOTE_LF   FLOAT
#define GPL_PROMOTE_LD   DOUBLE
#define GPL_PROMOTE_LC   FCOMPLEX
#define GPL_PROMOTE_LZ   DCOMPLEX

#define GPL_PROMOTE_FB   FLOAT
#define GPL_PROMOTE_FS   FLOAT
#define GPL_PROMOTE_FI   FLOAT
#define GPL_PROMOTE_FL   FLOAT
#define GPL_PROMOTE_FF   FLOAT
#define GPL_PROMOTE_FD   DOUBLE
#define GPL_PROMOTE_FC   FCOMPLEX
#define GPL_PROMOTE_FZ   DCOMPLEX

#define GPL_PROMOTE_DB   DOUBLE
#define GPL_PROMOTE_DS   DOUBLE
#define GPL_PROMOTE_DI   DOUBLE
#define GPL_PROMOTE_DL   DOUBLE
#define GPL_PROMOTE_DF   DOUBLE
#define GPL_PROMOTE_DD   DOUBLE
#define GPL_PROMOTE_DC   DCOMPLEX
#define GPL_PROMOTE_DZ   DCOMPLEX

#define GPL_PROMOTE_CB   FCOMPLEX
#define GPL_PROMOTE_CS   FCOMPLEX
#define GPL_PROMOTE_CI   FCOMPLEX
#define GPL_PROMOTE_CL   FCOMPLEX
#define GPL_PROMOTE_CF   FCOMPLEX
#define GPL_PROMOTE_CD   DCOMPLEX
#define GPL_PROMOTE_CC   FCOMPLEX
#define GPL_PROMOTE_CZ   DCOMPLEX

#define GPL_PROMOTE_ZB   DCOMPLEX
#define GPL_PROMOTE_ZS   DCOMPLEX
#define GPL_PROMOTE_ZI   DCOMPLEX
#define GPL_PROMOTE_ZL   DCOMPLEX
#define GPL_PROMOTE_ZF   DCOMPLEX
#define GPL_PROMOTE_ZD   DCOMPLEX
#define GPL_PROMOTE_ZC   DCOMPLEX
#define GPL_PROMOTE_ZZ   DCOMPLEX

#endif /* _GPL_MACROS_H -------------------------------------------------------*/
