/*
 * gpl_ops0.c --
 *
 *	Miscellaneous vector operations in GPLib.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 */
#include <math.h>
#include "gpl_ops0.h"

/*---------------------------------------------------------------------------*/
/* FILLING ARRAYS */

#undef FUNCTION
#define FUNCTION(SUFFIX, type_t)					      \
type_t *GPL_JOIN(gpl_fill_,SUFFIX)(type_t x[], int n, type_t value)		      \
{ int i; for (i=0 ; i<n ; ++i) x[i]= value; return x; }
FUNCTION(b, gpl_byte_t)
FUNCTION(s, short)
FUNCTION(i, int)
FUNCTION(l, long)
FUNCTION(f, float)
FUNCTION(d, double)

#undef FUNCTION
#define FUNCTION(SUFFIX, type_t, real_t)				      \
type_t *GPL_JOIN(gpl_fill_,SUFFIX)(type_t z[], int n, type_t value)		      \
{									      \
  int i;								      \
  real_t *x= (real_t *)z, re= value.re, im= value.im;			      \
  for (i=0, n*=2 ; i<n ; i+=2) {					      \
    x[i]= re;								      \
    x[i+1]= im;								      \
  }									      \
  return z;								      \
}
FUNCTION(c, gpl_fcomplex_t, float)
FUNCTION(z, gpl_dcomplex_t, double)

/*---------------------------------------------------------------------------*/
#undef FUNCTION
#define FUNCTION(SUFFIX, CTYPE)						      \
double GPL_JOIN(gpl_euclidian_norm_,SUFFIX)(const CTYPE x[], int number)	      \
{									      \
  int i; double sum2= 0.0, xmax= 0.0, xabs;				      \
  for (i=0 ; i<number ; i++) {						      \
    if ((xabs= x[i])<0.0) xabs= -xabs;					      \
    if (xabs>xmax) {							      \
      xmax/= xabs;							      \
      sum2= sum2*xmax*xmax + 1.0;					      \
      xmax= xabs;							      \
    } else if (xabs) {							      \
      xabs/= xmax;							      \
      sum2+= xabs*xabs;							      \
    }									      \
  }									      \
  return xmax*sqrt(sum2);						      \
}
#if 0
FUNCTION(b, gpl_byte_t)
FUNCTION(s, short)
FUNCTION(i, int)
FUNCTION(l, long)
#endif
FUNCTION(f, float)
FUNCTION(d, double)

/*---------------------------------------------------------------------------*/

#undef FUNCTION
#define FUNCTION(SUFFIX, CTYPE)						      \
double GPL_JOIN(gpl_rms_,SUFFIX)(const CTYPE x[], int number)		      \
{									      \
  int i; double sum2= 0.0, xmax= 0.0, xabs;				      \
  for (i=0 ; i<number ; i++) {						      \
    if ((xabs= x[i])<0) xabs= -xabs;					      \
    if (xabs>xmax) {							      \
      xmax/= xabs;							      \
      sum2= sum2*xmax*xmax + 1.0;					      \
      xmax= xabs;							      \
    } else if (xabs) {							      \
      xabs/= xmax;							      \
      sum2+= xabs*xabs;							      \
    }									      \
  }									      \
  return xmax*sqrt(sum2/number);					      \
}
FUNCTION(b, gpl_byte_t)
FUNCTION(s, short)
FUNCTION(i, int)
FUNCTION(l, long)
FUNCTION(f, float)
FUNCTION(d, double)

/*---------------------------------------------------------------------------*/

#undef FUNCTION
#define FUNCTION(SUFFIX, CTYPE)						      \
double GPL_JOIN(gpl_dot_product_,SUFFIX)(const CTYPE x[], const CTYPE y[],	      \
				     int number)			      \
{									      \
  int i; double dot_product;						      \
  for (i=1, dot_product= x[0]*y[0] ; i<number ; i++) {			      \
    dot_product+= x[i]*y[i];						      \
  }									      \
  return dot_product;							      \
}
#if 0
FUNCTION(b, gpl_byte_t)
FUNCTION(s, short)
FUNCTION(i, int)
FUNCTION(l, long)
#endif
FUNCTION(f, float)
FUNCTION(d, double)

/*---------------------------------------------------------------------------*/

#undef FUNCTION
#define FUNCTION(SUFFIX, CTYPE)						      \
CTYPE *GPL_JOIN(gpl_xpay_,SUFFIX)(CTYPE xpay[], const CTYPE x[], CTYPE a,	      \
			      const CTYPE y[], int number)		      \
{									      \
  int i;								      \
  									      \
  if (a==0) {								      \
    for (i=0 ; i<number ; i++) xpay[i]= x[i];				      \
  } else if (a==1) {							      \
    for (i=0 ; i<number ; i++) xpay[i]= x[i] + y[i];			      \
  } else {								      \
    for (i=0 ; i<number ; i++) xpay[i]= x[i] + a*y[i];			      \
  }									      \
  return xpay;								      \
}
#if 0
FUNCTION(b, gpl_byte_t)
FUNCTION(s, short)
FUNCTION(i, int)
FUNCTION(l, long)
#endif
FUNCTION(f, float)
FUNCTION(d, double)

/*---------------------------------------------------------------------------*/
/* AMAX: MAXIMUM VALUE OF ABS(X) */

#undef FUNCTION
#define FUNCTION(SUFFIX, CTYPE)						      \
CTYPE GPL_JOIN(gpl_amax_,SUFFIX)(const CTYPE x[], int number)		      \
{									      \
  int i; CTYPE amax, ax;						      \
  for (amax=0, i=0 ; i<number ; i++) {					      \
    if ((ax= x[i]) < 0) ax= -ax;					      \
    if (ax > amax) amax= ax;						      \
  }									      \
  return amax;								      \
}
#if 0
FUNCTION(b, gpl_byte_t)
FUNCTION(s, short)
FUNCTION(i, int)
FUNCTION(l, long)
#endif
FUNCTION(f, float)
FUNCTION(d, double)

/*---------------------------------------------------------------------------*/

#undef FUNCTION
#define FUNCTION(SUFFIX, CTYPE)						      \
CTYPE GPL_JOIN(gpl_min_,SUFFIX)(const CTYPE x[], int number)		      \
{									      \
  int i; CTYPE xmin= x[0];						      \
  for (i=1 ; i<number ; i++) if (x[i]<xmin) xmin= x[i];			      \
  return xmin;								      \
}									      \
CTYPE GPL_JOIN(gpl_max_,SUFFIX)(const CTYPE x[], int number)		      \
{									      \
  int i; CTYPE xmax= x[0];						      \
  for (i=1 ; i<number ; i++) if (x[i]>xmax) xmax= x[i];			      \
  return xmax;								      \
}
FUNCTION(b, gpl_byte_t)
FUNCTION(s, short)
FUNCTION(i, int)
FUNCTION(l, long)
FUNCTION(f, float)
FUNCTION(d, double)

/*---------------------------------------------------------------------------*/
#undef FUNCTION
#define FUNCTION(SUFFIX, CTYPE)						      \
CTYPE GPL_JOIN(gpl_sum_,SUFFIX)(const CTYPE x[], int number)		      \
{									      \
  int i; CTYPE sum= x[0];						      \
  for (i=1 ; i<number ; i++) sum+= x[i];				      \
  return sum;								      \
}
FUNCTION(b, gpl_byte_t)
FUNCTION(s, short)
FUNCTION(i, int)
FUNCTION(l, long)
FUNCTION(f, float)
FUNCTION(d, double)

#undef FUNCTION
#define FUNCTION(SUFFIX, CTYPE, RTYPE)					      \
void GPL_JOIN(gpl_sum_,SUFFIX)(CTYPE *sum, const CTYPE z[], int number)	      \
{									      \
  int i; const RTYPE *x= (RTYPE *)z; RTYPE sum_re= x[0], sum_im= x[1];	      \
  for (i=2, number*=2 ; i<number ; i+=2) {sum_re+= x[i]; sum_im+=x[i+1];}     \
  sum->re= sum_re; sum->im= sum_im;					      \
}
FUNCTION(c, gpl_fcomplex_t, float)
FUNCTION(z, gpl_dcomplex_t, double)
/*---------------------------------------------------------------------------*/
#undef FUNCTION
#define FUNCTION(SUFFIX, CTYPE)						      \
double GPL_JOIN(gpl_avg_,SUFFIX)(const CTYPE x[], int number)		      \
{									      \
  int i; double sum= x[0];						      \
  for (i=1 ; i<number ; i++) sum+= x[i];				      \
  return sum/(double)number;						      \
}
FUNCTION(b, gpl_byte_t)
FUNCTION(s, short)
FUNCTION(i, int)
FUNCTION(l, long)
FUNCTION(f, float)
FUNCTION(d, double)

#undef FUNCTION
#define FUNCTION(SUFFIX, CTYPE, RTYPE)					      \
void GPL_JOIN(gpl_avg_,SUFFIX)(CTYPE *avg, const CTYPE z[], int number)	      \
{									      \
  int i; const RTYPE *x= (RTYPE *)z; RTYPE sum_re= x[0], sum_im= x[1];	      \
  for (i=2, number*=2 ; i<number ; i+=2) {sum_re+= x[i]; sum_im+=x[i+1];}     \
  avg->re= sum_re/(RTYPE)number; avg->im= sum_im/(RTYPE)number;		      \
}
FUNCTION(c, gpl_fcomplex_t, float)
FUNCTION(z, gpl_dcomplex_t, double)
/*---------------------------------------------------------------------------*/
