/*
 * gpl_ops2.c --
 *
 *	Matrix oriented routines.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 */

#ifndef GPL_OPS2_C
#define GPL_OPS2_C 1
#include "gpl_base.h"
#include "gpl_ops2.h"

#define NAME(prefix,suffix) GPL_JOIN4(gpl_,prefix,_,suffix)
#define IS_COMPLEX 0

#define CTYPE  gpl_byte_t
#define SUFFIX b
#include __FILE__

#define CTYPE  short
#define SUFFIX s
#include __FILE__

#define CTYPE  int
#define SUFFIX i
#include __FILE__

#define CTYPE  long
#define SUFFIX l
#include __FILE__

#define CTYPE  float
#define SUFFIX f
#include __FILE__

#define CTYPE  double
#define SUFFIX d
#include __FILE__

#undef IS_COMPLEX
#define IS_COMPLEX 1

#define CTYPE  gpl_fcomplex_t
#define RTYPE  float
#define SUFFIX c
#include __FILE__

#define CTYPE  gpl_dcomplex_t
#define RTYPE  double
#define SUFFIX z
#include __FILE__

#else /* GPL_OPS2_C defined ---------------------------------------------------*/

#if IS_COMPLEX
# define RESET(c)        (c).re= (c).im= 0
# define MMLT_OP(c,a,b)  (c).re+= (a).re * (b).re - (a).im * (b).im; \
                         (c).im+= (a).re * (b).im - (a).im * (b).re
# define SET(d,c)        (d).re= (c).re; (d).im= (c).im
#else
# define RTYPE           CTYPE
# define RESET(c)        (c)=0
# define MMLT_OP(c,a,b)  (c)+= (a) * (b)
# define SET(d,c)        (d)= (c)
#endif

/*---------------------------------------------------------------------------*/
/* MATRIX MULTIPLICATION */
CTYPE *NAME(mmlt, SUFFIX)(CTYPE dst[], const CTYPE lop[], const CTYPE rop[],
			   int ni, int nj, int nk, int transpose)
{
  int i, j, k;
  CTYPE sum;
  transpose&= 3;
  if (transpose == 0) {
    for (i=0 ; i<ni ; i++) {
      for (j=0 ; j<nj ; j++) {
	for (RESET(sum), k=0 ; k<nk ; k++) {
	  MMLT_OP(sum, lop[i+ni*k], rop[k+nk*j]);
	}
	SET(dst[i+ni*j], sum);
      }
    }
  } else if (transpose == 1) {
    for (i=0 ; i<ni ; i++) {
      for (j=0 ; j<nj ; j++) {
	for (RESET(sum), k=0 ; k<nk ; k++) {
	  MMLT_OP(sum, lop[k+nk*i], rop[k+nk*j]);
	}
	SET(dst[i+ni*j], sum);
      }
    }
  } else if (transpose == 2) {
    for (i=0 ; i<ni ; i++) {
      for (j=0 ; j<nj ; j++) {
	for (RESET(sum), k=0 ; k<nk ; k++) {
	  MMLT_OP(sum, lop[i+ni*k], rop[j+nj*k]);
	}
	SET(dst[i+ni*j], sum);
      }
    }
  } else /* transpose = 3 */ {
    for (i=0 ; i<ni ; i++) {
      for (j=0 ; j<nj ; j++) {
	for (RESET(sum), k=0 ; k<nk ; k++) {
	  MMLT_OP(sum, lop[k+nk*i], rop[j+nj*k]);
	}
	SET(dst[i+ni*j], sum);
      }
    }
  }
  return dst;
}
/*---------------------------------------------------------------------------*/
/* MATRIX TRANSPOSE */
CTYPE *NAME(transpose, SUFFIX)(CTYPE dst[], const CTYPE src[],
			       int dim1, int dim2)
{
  int i1, i2, j;
  RTYPE *a= (RTYPE *)dst;
  const RTYPE *b= (const RTYPE *)src;
#if IS_COMPLEX
  if ((dim1*= 2) >= (dim2*= 2)) {
    for (i2=0 ; i2<dim2 ; i2+=2, a+=2, b+=dim1) {
      for (i1=0, j=0 ; i1<dim1 ; i1+=2, j+=dim2) {
	a[j]= b[i1];
	a[j+1]= b[i1+1];
      }
    }
  } else {
    for (i1=0 ; i1<dim1 ; i1+=2, a+=dim2, b+=2) {
      for (i2=0, j=0 ; i2<dim2 ; i2+=2, j+=dim1) {
	a[i2]= b[j];
	a[i2+1]= b[j+1];
      }
    }
  }
#else
  if (dim1 >= dim2) {
    for (i2=0 ; i2<dim2 ; i2++, a++, b+=dim1) {
      for (i1=0, j=0 ; i1<dim1 ; i1++, j+=dim2) {
	a[j]= b[i1];
      }
    }
  } else {
    for (i1=0 ; i1<dim1 ; i1++, a+=dim2, b++) {
      for (i2=0, j=0 ; i2<dim2 ; i2++, j+=dim1) {
	a[i2]= b[j];
      }
    }
  }
#endif
  return dst;
}

/*---------------------------------------------------------------------------*/
#undef CTYPE
#undef RTYPE
#undef SET
#undef MMLT_OP
#undef RESET
#undef SUFFIX
#endif /* GPL_OPS2_C ----------------------------------------------------------*/
