/*                                                                    -*- C -*-
 * machine.c -- generate machine dependant settings.
 *
 * Copyright (C) 2001 Eric Thibaut <thiebaut@obs.univ-lyon1.fr>
 *
 * History:
 *   $Id$
 *   $Log$
 */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

int main(int argc, char *argv[])
{
  const char *format= "s/@%s@/%s/g\n";
  int i, status= EXIT_SUCCESS;
  FILE *out= stdout;
  const char *sep= "%";
  const char *s_int16, *s_int32, *s_byteorder;
  unsigned long value;
  char *ptr;


  if (argc == 1) {
    out= stdout;
  } else if (argc == 2) {
    const char *filename= argv[1];
    if (access(filename, F_OK) == 0) {
      fprintf(stderr, "%s: file \"%s\" already exists, aborting...\n",
	      argv[0], filename);
      return EXIT_FAILURE;
    }
    if ((out= fopen(filename, "w")) == NULL) {
      fprintf(stderr, "%s: cannot open \"%s\", aborting...\n",
	      argv[0], filename);
      return EXIT_FAILURE;
    }
  } else {
    fprintf(stderr, "usage: %s [FILENAME]\n", argv[0]);
    return EXIT_FAILURE;
  }


  fprintf(stderr, "16-bit integer type on this machine: ");
  fflush(stderr);
#undef _
#define _(type) (sizeof(type)==2 && sizeof(unsigned type)==2) s_int16= #type
  if _(int);
  else if _(short);
  else if _(long);
  else {
    s_int16= "UNKNOWN";
    status=  EXIT_FAILURE;
  }
  fprintf(stderr, "%s\n", s_int16);
  fprintf(out, format, "int16", s_int16);


  fprintf(stderr, "32-bit integer type on this machine: ");
  fflush(stderr);
#undef _
#define _(type) (sizeof(type)==4 && sizeof(unsigned type)==4) s_int32= #type
  if _(int);
  else if _(long);
  else if _(short);
  else {
    s_int32= "UNKNOWN";
    status=  EXIT_FAILURE;
  }
  fprintf(stderr, "%s\n", s_int32);
  fprintf(out, format, "int32", s_int32);


  fprintf(stderr, "Byte order on this machine: ");
  fflush(stderr);
  value= 0UL;
  ptr= (char *)&value;
  for (i=0 ; i<sizeof(value) ; ++i) ptr[i]= 4 - i%4;
  value&= 0xffffffff;
  if (value == 0x01020304) {
    s_byteorder= "LITTLE_ENDIAN";
  } else if (value == 0x04030201) {
    s_byteorder= "BIG_ENDIAN";
  } else if (value == 0x03040102) {
    s_byteorder= "PDP_ENDIAN";
  } else {
    s_byteorder= "UNKNOWN";
    status=  EXIT_FAILURE;
  }
  fprintf(stderr, "%s\n", s_byteorder);
  fprintf(out, format, "BYTE_ORDER", s_byteorder);


  if (out != stdout) fclose(out);
  return status;
}
